@@ -28,7 +28,7 @@ dependencies using raw pip. The virtual environment is not created, but one can
 issue `python3 -m venv venv/ && . venv/bin/activate` to create one.
 """
 
-__version__ = "1.2.1"
+__version__ = "1.3.0"
 __author__ = "Fridolin Pokorny <fridex.devel@gmail.com>"
 __title__ = "micropipenv"
 
@@ -48,7 +48,7 @@ from pathlib import Path
 from urllib.parse import urlparse
 
 _LOGGER = logging.getLogger(__title__)
-_SUPPORTED_PIP_STR = ">=9,<=22.0.3"  # Respects requirement in setup.py and latest pip to release date.
+_SUPPORTED_PIP_STR = ">=9,<=22.0.4"  # Respects requirement in setup.py and latest pip to release date.
 
 try:
     from pip import __version__ as pip_version
@@ -295,6 +295,77 @@ def _compute_pipfile_hash(pipfile):  # type: (Dict[str, Any]) -> str
     return hashlib.sha256(content.encode("utf8")).hexdigest()
 
 
+def _compute_poetry_hash(pyproject):  # type: (MutableMapping[str, Any]) -> str
+    """Compute pyproject.toml hash based on poetry content."""
+    poetry_data = pyproject["tool"]["poetry"]
+    relevant_keys = ["dependencies", "dev-dependencies", "source", "extras"]
+    relevant_content = {key: poetry_data.get(key) for key in relevant_keys}
+
+    return hashlib.sha256(json.dumps(relevant_content, sort_keys=True).encode()).hexdigest()
+
+
+def _get_installed_python_version():  # type: () -> str
+    return "{}.{}".format(sys.version_info.major, sys.version_info.minor)
+
+
+def _validate_poetry_python_version(
+    poetry_lock, current_python, debug=False
+):  # type: (MutableMapping[str, Any], str, bool) -> None
+    # TODO: Implement or use external parser for Python versions in Poetry specification.
+    # See for details: https://github.com/thoth-station/micropipenv/issues/187
+    wanted_python_version = poetry_lock["metadata"]["python-versions"]
+    message = (
+        "Warning: Currently, Micropipenv is not able to parse complex Python version specifications used by Poetry. "
+        f"Desired version: {wanted_python_version}, current version: {current_python}."
+    )
+    level = "debug" if debug else "warning"
+    getattr(_LOGGER, level)(message)
+
+
+def verify_poetry_lockfile(
+    pyproject=None, poetry_lock=None, current_python_version=None
+):  # type: (Optional[MutableMapping[str, Any]], Optional[MutableMapping[str, Any]], Optional[str]) -> None
+    """Validate that Poetry.lock is up to date with pyproject.toml."""
+    if pyproject is None or poetry_lock is None:
+        poetry_lock, pyproject = _read_poetry()
+    if current_python_version is None:
+        current_python_version = _get_installed_python_version()
+    _validate_poetry_python_version(poetry_lock, current_python_version)
+    pyproject_hash = _compute_poetry_hash(pyproject)
+    poetry_lock_hash = poetry_lock["metadata"]["content-hash"]
+
+    if pyproject_hash != poetry_lock_hash:
+        raise HashMismatch(
+            "Poetry.lock hash {!r} does not correspond to hash computed based on "
+            "pyproject.toml {!r}, aborting deployment".format(poetry_lock_hash, pyproject_hash)
+        )
+
+
+def verify_pipenv_lockfile(
+    pipfile=None, pipfile_lock=None
+):  # type: (Optional[Dict[str, Any]], Optional[Dict[str, Any]]) -> None
+    """Validate that Pipfile.lock is up to date with Pipfile."""
+    pipfile_lock = pipfile_lock or _read_pipfile_lock()
+    pipenv_python_version = pipfile_lock["_meta"].get("requires", {}).get("python_version")
+    if pipenv_python_version is not None:
+        installed_python_version = _get_installed_python_version()
+        if pipenv_python_version != installed_python_version:
+            raise PythonVersionMismatch(
+                "Running Python version {}, but Pipfile.lock requires "
+                "Python version {}".format(installed_python_version, pipenv_python_version)
+            )
+    else:
+        _LOGGER.warning("No Python version requirement in Pipfile.lock found, no Python version check is performed")
+
+    pipfile_lock_hash = pipfile_lock.get("_meta", {}).get("hash", {}).get("sha256")
+    pipfile_hash = _compute_pipfile_hash(pipfile or _read_pipfile())
+    if pipfile_hash != pipfile_lock_hash:
+        raise HashMismatch(
+            "Pipfile.lock hash {!r} does not correspond to hash computed based on "
+            "Pipfile {!r}, aborting deployment".format(pipfile_lock_hash, pipfile_hash)
+        )
+
+
 def install_pipenv(
     pip_bin=_PIP_BIN, pipfile=None, pipfile_lock=None, *, deploy=False, dev=False, pip_args=None
 ):  # type: (str, Optional[Dict[str, Any]], Optional[Dict[str, Any]], bool, bool, Optional[List[str]]) -> None
@@ -304,23 +375,7 @@ def install_pipenv(
 
     sections = get_requirements_sections(pipfile_lock=pipfile_lock, no_dev=not dev)
     if deploy:
-        python_version = pipfile_lock["_meta"].get("requires", {}).get("python_version")
-        if python_version is not None:
-            if python_version != "{}.{}".format(sys.version_info.major, sys.version_info.minor):
-                raise PythonVersionMismatch(
-                    "Running Python version {}.{}, but Pipfile.lock requires "
-                    "Python version {}".format(sys.version_info.major, sys.version_info.minor, python_version)
-                )
-        else:
-            _LOGGER.warning("No Python version requirement in Pipfile.lock found, no Python version check is performed")
-
-        pipfile_lock_hash = pipfile_lock.get("_meta", {}).get("hash", {}).get("sha256")
-        pipfile_hash = _compute_pipfile_hash(pipfile or _read_pipfile())
-        if pipfile_hash != pipfile_lock_hash:
-            raise HashMismatch(
-                "Pipfile.lock hash {!r} does not correspond to hash computed based on "
-                "Pipfile {!r}, aborting deployment".format(pipfile_lock_hash, pipfile_hash)
-            )
+        verify_pipenv_lockfile(pipfile, pipfile_lock)
 
     tmp_file = tempfile.NamedTemporaryFile("w", prefix="requirements_micropipenv-", suffix=".txt", delete=False)
     _LOGGER.debug("Using temporary file for storing requirements: %r", tmp_file.name)
@@ -609,7 +664,7 @@ def _requirements2pipfile_lock(requirements_txt_path=None):  # type: (Optional[s
             "hash": {"sha256": requirements_hash},
             "pipfile-spec": 6,
             "sources": sources,
-            "requires": {"python_version": "{}.{}".format(sys.version_info.major, sys.version_info.minor)},
+            "requires": {"python_version": _get_installed_python_version()},
         },
         "default": result,
         "develop": {},
@@ -667,18 +722,11 @@ def _poetry2pipfile_lock(
     """Convert Poetry files to Pipfile.lock as Pipenv would produce."""
     poetry_lock, pyproject_toml = _read_poetry()
 
-    # If deploy is specified, show a message about the lack of Python versions
-    # parsing capabilities in Micropipenv.
-    # TODO: Implement or use external parser for Python versions in Poetry specification.
-    # See for details: https://github.com/thoth-station/micropipenv/issues/187
-    current_python_version = "{}.{}".format(sys.version_info.major, sys.version_info.minor)
-    wanted_python_version = poetry_lock["metadata"]["python-versions"]
-    message = (
-        "Warning: Currently, Micropipenv is not able to parse complex Python version specifications used by Poetry. "
-        f"Desired version: {wanted_python_version}, current version: {current_python_version}."
-    )
-    level = "warning" if deploy else "debug"
-    getattr(_LOGGER, level)(message)
+    current_python_version = _get_installed_python_version()
+    if deploy:
+        verify_poetry_lockfile(pyproject_toml, poetry_lock, current_python_version)
+    else:
+        _validate_poetry_python_version(poetry_lock, current_python_version, debug=True)
 
     pyproject_poetry_section = pyproject_toml.get("tool", {}).get("poetry", {})
 
@@ -704,7 +752,8 @@ def _poetry2pipfile_lock(
     if only_direct:
         if not no_default:
             for dependency_name, info in pyproject_poetry_section.get("dependencies", {}).items():
-                default[dependency_name] = _translate_poetry_dependency(info)
+                if dependency_name != "python":
+                    default[dependency_name] = _translate_poetry_dependency(info)
 
         if not no_dev:
             for dependency_name, info in pyproject_poetry_section.get("dev-dependencies", {}).items():
@@ -723,8 +772,16 @@ def _poetry2pipfile_lock(
 
     additional_markers = defaultdict(list)
     skip_all_markers = object()
+    add_to_dev = list()
 
     for entry in poetry_lock["package"]:
+
+        if entry["category"] not in ("dev", "main"):
+            message = ("Unknown category for package '{}': '{}'. Supported categories are 'dev' and 'main'.").format(
+                entry["name"], entry["category"]
+            )
+            raise PoetryError(message)
+
         hashes = []
         for file_entry in poetry_lock["metadata"]["files"][entry["name"]]:
             hashes.append(file_entry["hash"])
@@ -766,12 +823,36 @@ def _poetry2pipfile_lock(
             # If there are no additional markers, we have to have a record
             # that the dependency has to be installed unconditionaly and that
             # we have to skip all other additional markers.
+            # Also, we don't care about "extra" markers which are computed separatedly
+            # and are usually not combined with other markers.
             if dependency_name not in pyproject_poetry_section.get("dependencies", {}):
-                if isinstance(dependency_info, dict) and "markers" in dependency_info:
+                if (
+                    isinstance(dependency_info, dict)
+                    and "markers" in dependency_info
+                    and not dependency_info["markers"].startswith("extra")
+                ):
                     additional_markers[normalize_package_name(dependency_name)].append(dependency_info["markers"])
                 else:
                     additional_markers[normalize_package_name(dependency_name)].append(skip_all_markers)
 
+            # If package fits into both main and dev categories, poetry add it to the main one.
+            # This might be a problem in the following scenario:
+            # Let's say our project looks like this:
+            # dependencies:
+            # - A
+            # dev-dependencies:
+            # - B
+            #   - A (A is also a dependency of B)
+            # Package A is therefore in the main category so if we use "--do-default" we get
+            # only package B but no package A. But using requirements file requires to have
+            # all dependencies with their hashes and pinned versions.
+            # So, if a package is in dev and has a dependency in main, add the dependency also to dev or
+            # if a package is already in "add_to_dev", add there also all its dependencies.
+            if (
+                entry["category"] == "dev" and dependency_name in pyproject_poetry_section.get("dependencies", ())
+            ) or entry["name"] in add_to_dev:
+                add_to_dev.append(dependency_name)
+
         for extra_name, extras_listed in entry.get("extras", {}).items():
             # Turn requirement specification into the actual requirement name.
             all_extra_dependencies = set(Requirement(r.split(" ", maxsplit=1)[0]).name for r in extras_listed)
@@ -785,14 +866,13 @@ def _poetry2pipfile_lock(
         if "extras" in requirement:
             requirement["extras"] = sorted(requirement["extras"])
 
-        if entry["category"] == "main":
-            if not no_default:
-                default[entry["name"]] = requirement
-        elif entry["category"] == "dev":
-            if not no_dev:
-                develop[entry["name"]] = requirement
-        else:
-            raise PoetryError("Unknown category for package {}: {}".format(entry["name"], entry["category"]))
+        if entry["category"] == "main" and not no_default:
+            default[entry["name"]] = requirement
+
+        if (entry["category"] == "dev" and not no_dev) or (
+            entry["name"] in add_to_dev and entry["name"] not in default
+        ):
+            develop[entry["name"]] = requirement
 
     for dependency_name, markers in additional_markers.items():
         # If a package depends on another package unconditionaly
@@ -828,7 +908,7 @@ def _poetry2pipfile_lock(
             "hash": {"sha256": poetry_lock["metadata"]["content-hash"]},
             "pipfile-spec": 6,
             "sources": sources,
-            "requires": {"python_version": "{}.{}".format(sys.version_info.major, sys.version_info.minor)},
+            "requires": {"python_version": _get_installed_python_version()},
         },
         "default": default,
         "develop": develop,
@@ -904,6 +984,24 @@ def method_discovery(ignore_files=None):  # type: (Optional[Sequence[str]]) -> s
     return _FILE_METHOD_MAP[longest_path.name]
 
 
+def verify(method=None):  # type: (Optional[str]) -> None
+    """Check the lockfile to ensure it is up to date with the requirements file."""
+    if method is None:
+        method = method_discovery()
+
+    if method == "pipenv":
+        pipfile = _read_pipfile()
+        pipfile_lock = _read_pipfile_lock()
+        verify_pipenv_lockfile(pipfile, pipfile_lock)
+        return
+    elif method == "poetry":
+        poetry_lock, pyproject = _read_poetry()
+        verify_poetry_lockfile(pyproject, poetry_lock)
+        return
+
+    raise MicropipenvException("Unhandled method for checking lockfile: {}".format(method))
+
+
 def install(
     method=None, *, pip_bin=_PIP_BIN, deploy=False, dev=False, pip_args=None
 ):  # type: (Optional[str], str, bool, bool, Optional[List[str]]) -> None
@@ -997,8 +1095,11 @@ def get_requirements_sections(
 
 def _get_package_entry_str(
     package_name, info, *, no_hashes=False, no_versions=False
-):  # type: (str, Dict[str, Any], bool, bool) -> str
+):  # type: (str, Union[Dict[str, Any], str], bool, bool) -> str
     """Print entry for the given package."""
+    if isinstance(info, str):
+        return package_name + info + "\n"
+
     if "git" in info:  # A special case for a VCS package
         if info.get("editable", False):
             result = "--editable git+{}".format(info["git"])
@@ -1196,6 +1297,15 @@ def main(argv=None):  # type: (Optional[List[str]]) -> int
 
     subparsers = parser.add_subparsers()
 
+    parser_verify = subparsers.add_parser("verify", help=verify.__doc__)
+    parser_verify.add_argument(
+        "--method",
+        help="Source of packages for the installation, perform detection if not provided.",
+        choices=["pipenv", "poetry"],
+        default=os.getenv("MICROPIPENV_METHOD"),
+    )
+    parser_verify.set_defaults(func=verify)
+
     parser_install = subparsers.add_parser("install", help=install.__doc__)
     parser_install.add_argument(
         "--deploy",
